#!/usr/bin/env python3

import os
import sys
import re
import hashlib

print("testing...") or sys.stdout.flush()
########################################################################
import blst

msg = b"assertion"  # this what we're signing
DST = b"MY-DST"  # domain separation tag

SK = blst.SecretKey()
SK.keygen(b"*" * 32)  # secret key

########################################################################
# generate public key and signature

pk_for_wire = blst.P1(SK).serialize()

#                               # optional vvvvvvvvvvv augmentation
sig_for_wire = blst.P2().hash_to(msg, DST, pk_for_wire).sign_with(SK).serialize()

########################################################################
# at this point 'pk_for_wire', 'sig_for_wire' and 'msg' are
# "sent over network," so now on "receiver" side

sig = blst.P2_Affine(sig_for_wire)
pk = blst.P1_Affine(pk_for_wire)
if not pk.in_group():  # vet the public key
    raise AssertionError("disaster")
ctx = blst.Pairing(True, DST)
ctx.aggregate(pk, sig, msg, pk_for_wire)
ctx.commit()
if not ctx.finalverify():
    raise AssertionError("disaster")

########################################################################
# generate public key and signature

pk_for_wire = blst.P2(SK).serialize()

#                               # optional vvvvvvvvvvv augmentation
sig_for_wire = blst.P1().hash_to(msg, DST, pk_for_wire).sign_with(SK).serialize()

########################################################################
# at this point 'pk_for_wire', 'sig_for_wire' and 'msg' are
# "sent over network," so now on "receiver" side

sig = blst.P1_Affine(sig_for_wire)
pk = blst.P2_Affine(pk_for_wire)
if not pk.in_group():  # vet the public key
    raise AssertionError("disaster")
ctx = blst.Pairing(True, DST)
ctx.aggregate(pk, sig, msg, pk_for_wire)
ctx.commit()
if not ctx.finalverify():
    raise AssertionError("disaster")

if sys.version_info.major < 3:
    print("OK")
    sys.exit(0)

########################################################################
# from https://github.com/supranational/blst/issues/5

pk_for_wire = bytes.fromhex(
    "ab10fc693d038b73d67279127501a05f0072cbb7147c68650ef6ac4e0a413e5cabd1f35c8711e1f7d9d885bbc3b8eddc"
)
sig_for_wire = bytes.fromhex(
    "a44158c08c8c584477770feec2afa24d5a0b0bab2800414cb9efbb37c40339b6318c9349dad8de27ae644376d71232580ff5102c7a8579a6d2627c6e40b0ced737a60c66c7ebd377c04bf5ac957bf05bc8b6b09fbd7bdd2a7fa1090b5a0760bb"
)
msg = bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000000")
DST = bytes.fromhex(
    "424c535f5349475f424c53313233383147325f584d443a5348412d3235365f535357555f524f5f504f505f"
)

sig = blst.P2_Affine(sig_for_wire)
pk = blst.P1_Affine(pk_for_wire)
if not pk.in_group():  # vet the public key
    raise AssertionError("disaster")
if sig.core_verify(pk, True, msg, DST) != blst.BLST_SUCCESS:
    raise AssertionError("disaster")

########################################################################
# test vectors from draft-irtf-cfrg-hash-to-curve

try:
    import json
    import glob

    coord = re.compile(r"0x([0-9a-f]+)(?:\s*,\s*0x([0-9a-f]+))?", re.IGNORECASE)

    def serialize_json_point(P):
        ret = b""
        x = coord.search(P["x"])
        if x.group(2):
            ret += bytes.fromhex(x.group(2))
        ret += bytes.fromhex(x.group(1))
        y = coord.search(P["y"])
        if y.group(2):
            ret += bytes.fromhex(y.group(2))
        ret += bytes.fromhex(y.group(1))
        return ret

    root_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../")
    vector_files_path = os.path.join(
        root_dir, "blst/bindings/vectors/hash_to_curve/*.json"
    )
    print("Looking into vector files at", vector_files_path)

    vector_files = glob.glob(vector_files_path)
    assert len(vector_files) == 4
    for file in vector_files:
        print(file)
        data = json.load(open(file))

        if data["curve"] == "BLS12-381 G1":
            point = blst.P1()
        else:
            point = blst.P2()

        DST = bytes(data["dst"], "ascii")

        if data["randomOracle"]:
            func = "point.hash_to(msg, DST).serialize()"
        else:
            func = "point.encode_to(msg, DST).serialize()"

        for vec in data["vectors"]:
            msg = bytes(vec["msg"], "ascii")
            if eval(func) != serialize_json_point(vec["P"]):
                raise AssertionError(msg)
except ModuleNotFoundError:
    print("skipping hash-to-curve KATs")

########################################################################
# test multi-scalar multiplication for self-consistency

points = []
scalars = []
total = 0
for _ in range(0, 42):
    p = os.urandom(8)
    s = int.from_bytes(os.urandom(8), "big")
    points.append(blst.G1().mult(p))
    scalars.append(s)
    total += s * int.from_bytes(p, "little")
a = blst.P1s.mult_pippenger(blst.P1s.to_affine(points), scalars)
if not a.is_equal(blst.G1().mult(total)):
    raise AssertionError("disaster")

points = []
scalars = []
total = 0
for _ in range(0, 42):
    p = os.urandom(8)
    s = int.from_bytes(os.urandom(8), "big")
    points.append(blst.G2().mult(p))
    scalars.append(s)
    total += s * int.from_bytes(p, "little")
a = blst.P2s.mult_pippenger(blst.P2s.to_affine(points), scalars)
if not a.is_equal(blst.G2().mult(total)):
    raise AssertionError("disaster")

########################################################################
# rudimentary blind signature PoC

# Signer's public key, implicitly trusted.
PK = blst.P1(SK).to_affine()

# User wants to have |msg| signed,
H_msg = blst.P2().hash_to(msg, DST)
# chooses random |r|,
r = blst.Scalar().from_bendian(os.urandom(32))  # should be PRF in real life...
# blinds the H(|msg|) with |r| and sends it to the Signer.
msg_for_wire = H_msg.dup().sign_with(r).serialize()

# Signer signs and sends the result back to the User.
blind_msg = blst.P2(msg_for_wire)
if not blind_msg.in_group():  # is User messing with Signer?
    raise AssertionError("disaster")
sig_for_wire = blind_msg.sign_with(SK).serialize()

# User ...
signature = blst.P2(sig_for_wire)
if not signature.in_group():  # is Signer messing with User?
    raise AssertionError("disaster")
# unblinds the result with 1/|r| to produce the actual |signature|,
signature = signature.sign_with(r.inverse()).to_affine()
# and double-checks if the Signer was honest?
C1 = blst.PT(signature)
C2 = blst.PT(H_msg.to_affine(), PK)
if not blst.PT.finalverify(C1, C2):
    raise AssertionError("disaster")

# Now |signature| can be verified as any other, e.g....
ctx = blst.Pairing(True, DST)
ctx.aggregate(PK, signature, msg)
ctx.commit()
if not ctx.finalverify():
    raise AssertionError("disaster")

########################################################################
# [a variant of] BBS+ PoC, https://eprint.iacr.org/2008/136.pdf

# Signer's public key, implicitly trusted.
w = blst.P2(SK)
w_for_wire = w.compress()

### The block of messages

msgs = ["assertion1", "assertion2", "assertion3", "assertionN"]
m = []
for msg in msgs:
    h = hashlib.sha512()
    h.update(msg.encode("utf-8"))
    m.append(blst.Scalar().from_bendian(h.digest()))

### Everybody involved:

count = len(msgs)
g = []
for i in range(count + 1):
    g.append(
        blst.P1().hash_to(
            b"\0" + i.to_bytes(4, "big") + b"\0" + count.to_bytes(4, "big"),
            DST,
            w_for_wire,
        )
    )

### Signing the block of messages.

e = blst.Scalar().from_bendian(os.urandom(32))  # should be PRF in real life...
s = blst.Scalar().from_bendian(os.urandom(32))

a = blst.G1().add(g[0].dup().mult(s))
for i in range(count):
    a.add(g[i + 1].dup().mult(m[i]))

A = a.dup().sign_with(e.dup().add(SK).inverse()).to_affine()

signature = [A, e, s]  # serialize to application liking

### Signature Verification.
# Verifier deserializes |signature|, recalculates |a| and verifies.
A, e, s = signature  # fake deserialization
a = blst.G1().add(g[0].dup().mult(s))
for i in range(count):
    a.add(g[i + 1].dup().mult(m[i]))

if not A.in_group():
    AssertionError("disaster")
C1 = blst.PT(blst.G2().mult(e).add(w).to_affine(), A)
C2 = blst.PT(a.to_affine())
if not blst.PT.finalverify(C1, C2):
    raise AssertionError("disaster")

### Blind-signing Committed Block of Messages.

# User creates |commitment| and sends it to Signer.
s_prime = blst.Scalar().from_bendian(os.urandom(32))
commitment = g[0].dup().sign_with(s_prime)
for i in range(count):
    commitment.add(g[i + 1].dup().mult(m[i]))

# Signer challenges User with |c|.
c = blst.Scalar().from_bendian(os.urandom(32))
# User provides proof of commitment.
s_tilde = blst.Scalar().from_bendian(os.urandom(32))
ZK = g[0].dup().sign_with(s_tilde)
r = [s_tilde.add(c.dup().mul(s_prime))]
for i in range(count):
    r.append(blst.Scalar().from_bendian(os.urandom(32)))
    ZK.add(g[i + 1].dup().mult(r[-1]))
    r[-1].add(c.dup().mul(m[i]))
# Signer verifies the proof, |ZK| and |r|.
ZK_verify = commitment.dup().neg().mult(c)
for i in range(count + 1):
    ZK_verify.add(g[i].dup().mult(r[i]))
if not ZK_verify.is_equal(ZK):
    raise AssertionError("disaster")

# Signer signs the |commitment| and sends the |blind_signature| back.
if not commitment.in_group():  # is User messing with Signer?
    raise AssertionError("disaster")
e = blst.Scalar().from_bendian(os.urandom(32))
s_pprime = blst.Scalar().from_bendian(os.urandom(32))

A = (
    blst.G1()
    .add(g[0].dup().mult(s_pprime))
    .add(commitment)
    .sign_with(e.dup().add(SK).inverse())
    .to_affine()
)

blind_signature = [A, e, s_pprime]  # serialize to application liking

# User unblinds the |blind_signature| by adding |s_prime| and |s_pprime|.
signature = blind_signature.copy()  # fake deserialization
if not signature[0].in_group():  # is Signer messing with User?
    raise AssertionError("disaster")
signature[2].add(s_prime)  # serialize to application liking

### Signature Verification.
# Verifier deserializes |signature|, recalculates |a| and verifies.
A, e, s = signature  # fake deserialization
a = blst.G1().add(g[0].dup().mult(s))
for i in range(count):
    a.add(g[i + 1].dup().mult(m[i]))

if not A.in_group():
    raise AssertionError("disaster")
C1 = blst.PT(blst.G2().mult(e).add(w).to_affine(), A)
C2 = blst.PT(a.to_affine())
if not blst.PT.finalverify(C1, C2):
    raise AssertionError("disaster")

########################################################################
# low-order points

p11 = blst.P1(
    bytes.fromhex(
        "80803f0d09fec09a95f2ee7495323c15c162270c7cceaffa8566e941c66bcf206e72955d58b3b32e564de3209d672ca5"
    )
)
if p11.in_group():
    raise AssertionError("disaster")
if not p11.mult(11).is_inf():
    raise AssertionError("disaster")

p13 = blst.P2(
    bytes.fromhex(
        "808514e6f6c41ddb685cd8e351d7cb2ee891d54d5850d42c440365b15dd8ab06b53da9ebb22bc31cea2234d30499b09f01b290cae398ecc1fda3d4beb248af58bd351c3d99a5b6c770fdb54bf46b123261b55cf0762aeef2341e35e90608fc31"
    )
)
if p13.in_group():
    raise AssertionError("disaster")
if not p13.mult(13).is_inf():
    raise AssertionError("disaster")

p23 = blst.P2(
    bytes.fromhex(
        "acaaaa268d201642af7eeb2c46b03ecf8af71b902f652884577e52994047fcde2f8b5d931fdacb2575937b72ef2d3b0c0da315f7c31904614d8110d493d1d7a00da97d5b640be19e4cc5dd302ad8e17aa853035d7a56e5c24347164c186d6d00"
    )
)
if p23.in_group():
    raise AssertionError("disaster")
if not p23.mult(23).is_inf():
    raise AssertionError("disaster")

print("OK")



from ssz import (
    Serializable,
    uint64,
    bytes4,
    bytes32,
    bytes48,
)


ZERO_BYTES32 = b"\x00" * 32

# Eth2-spec constants taken from https://github.com/ethereum/eth2.0-specs/blob/dev/specs/phase0/beacon-chain.md
DOMAIN_DEPOSIT = bytes.fromhex("03000000")


class SigningData(Serializable):
    fields = [("object_root", bytes32), ("domain", bytes32)]


class ForkData(Serializable):
    fields = [
        ("current_version", bytes4),
        ("genesis_validators_root", bytes32),
    ]


class DepositMessage(Serializable):
    """
    Ref: https://github.com/ethereum/eth2.0-specs/blob/dev/specs/phase0/beacon-chain.md#depositmessage
    """

    fields = [
        ("pubkey", bytes48),
        ("withdrawal_credentials", bytes32),
        ("amount", uint64),
    ]


def compute_deposit_domain(fork_version: bytes) -> bytes:
    """
    Deposit-only `compute_domain`
    """
    if len(fork_version) != 4:
        raise ValueError(f"Fork version should be in 4 bytes. Got {len(fork_version)}.")
    domain_type = DOMAIN_DEPOSIT
    fork_data_root = compute_deposit_fork_data_root(fork_version)
    return domain_type + fork_data_root[:28]


def compute_deposit_fork_data_root(current_version: bytes) -> bytes:
    """
    Return the appropriate ForkData root for a given deposit version.
    """
    genesis_validators_root = ZERO_BYTES32  # For deposit, it's fixed value
    if len(current_version) != 4:
        raise ValueError(
            f"Fork version should be in 4 bytes. Got {len(current_version)}."
        )
    return ForkData(
        current_version=current_version,
        genesis_validators_root=genesis_validators_root,
    ).hash_tree_root


def compute_signing_root(ssz_object: Serializable, domain: bytes) -> bytes:
    """
    Return the signing root of an object by calculating the root of the object-domain tree.
    The root is the hash tree root of:
    https://github.com/ethereum/eth2.0-specs/blob/dev/specs/phase0/beacon-chain.md#signingdata
    """
    if len(domain) != 32:
        raise ValueError(f"Domain should be in 32 bytes. Got {len(domain)}.")
    domain_wrapped_object = SigningData(
        object_root=ssz_object.hash_tree_root,
        domain=domain,
    )
    return domain_wrapped_object.hash_tree_root



from typing import List, Dict

from multicall import Call
from web3 import Web3

from lido_sdk.multicall import Multicall


def execute_contract_multicall(
    w3: Web3,
    registry_address: str,
    abi_method_name: str,
    abi_input: List[Dict],
    abi_returns: List[Dict],
    args_list: List[List],
):
    """
    @param w3: Web3 instance
    @param registry_address: Contract address in current network
    @param abi_method_name: Contract name of method
    @param abi_input: Params that this method receives, exp: [{'type': 'uint256', 'name': 'contractId'}, ...]
    @param abi_returns: Params that function returns, exp: [{'type': 'uint256', 'name': 'signedKeys'}, ...]
    @param args_list: List of bunches of arg each of those will be used to call contract, exp: [[True, 1], [True, 2], ...]
    @return: List of results from each call that we did.
    """
    return Multicall(
        calls=[
            _create_contract_call(
                w3, registry_address, abi_method_name, abi_input, abi_returns, args
            )
            for args in args_list
        ],
        _w3=w3,
    )()


def execute_contract_call(
    w3: Web3,
    registry_address: str,
    abi_method_name: str,
    abi_input: List[Dict],
    abi_returns: List[Dict],
    args: List = None,
) -> Dict:
    """
    @param w3: Web3 instance
    @param registry_address: Contract address in current network
    @param abi_method_name: Contract name of method
    @param abi_input: Params that this method receives, exp: [{'type': 'uint256', 'name': 'contractId'}, ...]
    @param abi_returns: Params that function returns, exp: [{'type': 'uint256', 'name': 'signedKeys'}, ...]
    @param args: List of arguments that will be used to call this function
    @return: List of results from each call that we did.
    """
    return _create_contract_call(
        w3, registry_address, abi_method_name, abi_input, abi_returns, args
    )()


def _create_contract_call(
    w3: Web3,
    registry_address: str,
    abi_method_name: str,
    abi_input: List[Dict],
    abi_returns: List[Dict],
    args: List = None,
) -> Call:
    input_args = ",".join([x["type"] for x in abi_input])
    output_args = ",".join([x["type"] for x in abi_returns])
    returns = [(abi_return["name"], lambda y: y) for abi_return in abi_returns]

    function = f"{abi_method_name}({input_args})({output_args})"

    call = Call(
        target=registry_address,
        function=function,
        returns=returns,
        _w3=w3,
    )

    call.args = args

    return call



import json
import os

from lido_sdk.contract.contract import Contract
from lido_sdk.network import Network


LIDO_ADDRESSES = {
    Network.Mainnet: "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84",
    Network.Görli: "0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F",
    Network.Ropsten: "0xd40EefCFaB888C9159a61221def03bF77773FC19",
}

NODE_OPS_ADDRESSES = {
    Network.Mainnet: "0x55032650b14df07b85bF18A3a3eC8E0Af2e028d5",
    Network.Görli: "0x9D4AF1Ee19Dad8857db3a45B0374c81c8A1C6320",
    Network.Ropsten: "0x32c6f34F3920E8c0074241619c02be2fB722a68d",
}


def _get_contract_abi(contract_name: str):
    script_dir = os.path.dirname(__file__)

    with open(os.path.join(script_dir, "abi", contract_name)) as file:
        return json.load(file)


# Load all supported contracts here
LidoContract = Contract(LIDO_ADDRESSES, _get_contract_abi("Lido.json"))
NodeOpsContract = Contract(
    NODE_OPS_ADDRESSES, _get_contract_abi("NodeOperatorsRegistry.json")
)



from typing import Dict, List, Optional

from web3 import Web3


class Contract:
    """
    Contract class that converts abi interface and list of registry addresses to useful object.
    This is some kind of simplified interfaces from brownie.

    When you receive the Contract instance you will be able to use all contract methods in simple and multicall way.
    Example:
        LidoContract.{contractMethodName}(web3, args)
        or for multicall
        LidoContract.{contractMethodName}_multicall(web3, [args0, args1, ...])
        it is similar to
        LidoContract.{contractMethodName}(web3, args0)
        LidoContract.{contractMethodName}(web3, args1)
        ...
    """

    def __init__(self, registry_addresses: Dict[int, str], contract_abi: List[Dict]):
        """
        @param registry_addresses: It is a dictionary where chain_id is a key and str is the address in this Chain
        where contract is deployed.
        @param contract_abi: Typical contract interface

        Example of usage can be found in lido/contract/load_contract.py
        """
        self.registry_addresses = registry_addresses
        self.contract_abi = contract_abi

        for abi_element in contract_abi:
            if abi_element["type"] == "function":
                self._create_contract_method(abi_element)

    def _create_contract_method(self, abi_function):
        """Create all methods announced in contract's abi"""

        def call(w3: Web3, args: Optional[List] = None):
            from lido_sdk.contract.execute_contract import execute_contract_call

            return execute_contract_call(
                w3,
                self.registry_addresses[w3.eth.chain_id],
                abi_function["name"],
                abi_function["inputs"],
                abi_function["outputs"],
                args=args,
            )

        def multicall(w3: Web3, args_list: Optional[List[List]] = None):
            from lido_sdk.contract.execute_contract import execute_contract_multicall

            args_list = args_list or [[]]

            return execute_contract_multicall(
                w3,
                self.registry_addresses[w3.eth.chain_id],
                abi_function["name"],
                abi_function["inputs"],
                abi_function["outputs"],
                args_list=args_list,
            )

        setattr(self, abi_function["name"], call)
        setattr(self, f"{abi_function['name']}_multicall", multicall)



from typing import List, Optional, Tuple, Dict

from web3 import Web3

from lido_sdk import config
from lido_sdk.methods import (
    find_duplicated_keys,
    validate_keys,
    get_operators_indexes,
    get_operators_data,
    get_operators_keys,
    get_status,
)
from lido_sdk.methods.typing import Operator, OperatorKey
from lido_sdk.network import Network


class LidoException(Exception):
    pass


class Lido:
    operators_indexes = None
    operators = None
    keys = None

    def __init__(self, w3: Web3, **kwargs):
        self._w3 = w3

        self._set_configs(kwargs)

        if w3.eth.chain_id == Network.Görli:
            from web3.middleware import geth_poa_middleware

            # Checking by value b/c we don't know the key
            if geth_poa_middleware not in self._w3.middleware_onion:
                raise LidoException(
                    "PoA middleware isn't injected into Web3 middleware onion"
                )

    def _set_configs(self, kwargs: Dict):
        for key, value in kwargs.items():
            setattr(config, key, value)

    def get_operators_indexes(self) -> List[int]:
        """
        @return: List of operator's indexes in Lido.
        """
        self.operators_indexes = get_operators_indexes(self._w3)
        return self.operators_indexes

    def get_operators_data(
        self, operators_indexes: Optional[List[int]] = None
    ) -> List[Operator]:
        """
        It will fetch details for each operator specified.

        @param operators_indexes: List operators indexes to fetch.
        @return: List of operators details.
        """
        operators_indexes = operators_indexes or self.operators_indexes
        if operators_indexes is None:
            raise LidoException(
                "`get_operators_indexes` should be called first or provide `operators_indexes` param"
            )

        self.operators = get_operators_data(self._w3, operators_indexes)

        return self.operators

    def get_operators_keys(
        self, operators: Optional[List[Operator]] = None
    ) -> List[OperatorKey]:
        """
        Returns all keys for specified operators.

        @param operators: List of operators details. We need few fields to fetch "index" and "totalSigningKeys".
        @return: List of keys. Each key can be identified and refetched by "index" and "operator_index".
        """
        operators = operators or self.operators
        if operators is None:
            raise LidoException(
                "`get_operators_data` should be called first or provide `operators` param"
            )

        self.keys = get_operators_keys(self._w3, operators)

        return self.keys

    def validate_keys(
        self, keys: Optional[List[OperatorKey]] = None, strict: bool = False
    ) -> List[OperatorKey]:
        """
        Validate all provided keys with pub_key, signature, withdrawal_credentials and DepositDomain.

        @param keys: List of operators keys.
        @param strict: If strict is on, only actual contract withdraw credentials will be used check key.
        If not we will try old withdrawal credentials.
        @return: All invalid keys that were found.
        """
        keys = keys or self.keys
        if keys is None:
            raise LidoException(
                "`get_operators_keys` should be called first or provide `keys` param"
            )

        return validate_keys(self._w3, keys, strict)

    def find_duplicated_keys(
        self, keys: Optional[List[OperatorKey]] = None
    ) -> List[Tuple[OperatorKey, OperatorKey]]:
        """
        Find and returns all keys duplicates.

        @param keys: List a keys to check.
        @return: List of duplicate pairs keys.
        """
        keys = keys or self.keys
        if keys is None:
            raise LidoException(
                "`get_operators_keys` should be called first or provide `keys` param"
            )

        return find_duplicated_keys(keys)

    def get_status(self) -> dict:
        """
        Return a dict with Lido contract actual state.
        """
        return get_status(self._w3)

    def fetch_all_keys_and_validate(self) -> Dict[str, list]:
        """
        Function that makes all flow: fetches operators and keys, than it validate all keys.

        @return: Is a dict with two keys.
        - invalid_keys - for details see Lido.validate_keys method.
        - duplicated_keys - for details see Lido.find_duplicated_keys method.
        """
        self.get_operators_indexes()
        self.get_operators_data()
        self.get_operators_keys()
        invalid_keys = self.validate_keys()
        duplicated_keys = self.find_duplicated_keys()

        return {
            "invalid_keys": invalid_keys,
            "duplicated_keys": duplicated_keys,
        }



from typing import List

from web3 import Web3

from lido_sdk.contract.load_contract import NodeOpsContract
from lido_sdk.methods.typing import Operator, OperatorKey


def get_operators_indexes(w3: Web3) -> List[int]:
    """
    @param w3: Web3 instance
    @return: Node operators count
    """
    operators_count = NodeOpsContract.getNodeOperatorsCount(w3)[""]
    return [x for x in range(operators_count)]


def get_operators_data(w3: Web3, operators_index_list: List[int]) -> List[Operator]:
    """
    @param w3: Web3 instance
    @param operators_index_list: Operator's indexes to fetch
    @return: List of dictionary with operators details
    """
    operators = NodeOpsContract.getNodeOperator_multicall(
        w3,
        [(i, True) for i in operators_index_list],
    )

    # Add index to each operator
    for index, operator in zip(operators_index_list, operators):
        operator["index"] = index

    return operators


def get_operators_keys(w3: Web3, operators: List[Operator]) -> List[OperatorKey]:
    """
    @param w3: Web3 instance
    @param operators: List of method's details from get_operators_data. But we need only `index` and `totalSigningKeys`.
    @return: List of dicts (OperatorKey)
    """
    args_list = []

    for args in _index_generator(operators):
        args_list.append(args)

    keys = NodeOpsContract.getSigningKey_multicall(w3, args_list)

    for key, (operator_index, key_index) in zip(keys, _index_generator(operators)):
        key["index"] = key_index
        key["operator_index"] = operator_index

    return keys


def _index_generator(operators):
    for operator in operators:
        for key_index in range(operator["totalSigningKeys"]):
            yield operator["index"], key_index



from concurrent.futures import ThreadPoolExecutor

from multicall import Call, Multicall as DefaultMulticall

from lido_sdk import config
from lido_sdk.multicall.multicall_address import MULTICALL_ADDRESSES


class Multicall(DefaultMulticall):
    """
    Upgraded version of Multicall from https://github.com/banteg/multicall.py

    Improves:
        - Added MAX_CALLS_PER_MULTICALL param to avoid huge and slow batches in Multicall
        - results from multicall is not a dict, but a list now. We are making a lot of requests to one contract's method,
        so we don't wanna loose data.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.max_call_bunch = config.MULTICALL_MAX_BUNCH
        self.max_workers = config.MULTICALL_MAX_WORKERS
        self.max_retries = config.MULTICALL_MAX_RETRIES

    def __call__(self):
        calls_list = [
            self.calls[i : i + self.max_call_bunch]
            for i in range(0, len(self.calls), self.max_call_bunch)
        ]

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            thread_results = executor.map(self.execute, calls_list)

        result = []

        for thread_result in thread_results:
            result.extend(thread_result)

        return result

    def execute(self, calls):
        aggregate = Call(
            MULTICALL_ADDRESSES[self.w3.eth.chain_id],
            "aggregate((address,bytes)[])(uint256,bytes[])",
            returns=None,
            _w3=self.w3,
            block_id=self.block_id,
        )

        args = [[[call.target, call.data] for call in calls]]

        for retry_num in range(self.max_retries):
            try:
                block, outputs = aggregate(args)
            except ValueError as error:
                if retry_num == self.max_retries - 1:
                    raise error
            else:
                results = []
                for call, output in zip(self.calls, outputs):
                    results.append(call.decode_output(output))

                return results

        # Not expected exception
        raise Exception("Bug in Multicall")



from lido_sdk.methods.keys import find_duplicated_keys, validate_keys, validate_key
from lido_sdk.methods.operators import (
    get_operators_indexes,
    get_operators_data,
    get_operators_keys,
)
from lido_sdk.methods.stats import get_status



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 4.0.1
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _blst
else:
    import _blst

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

_swig_new_instance_method = _blst.SWIG_PyInstanceMethod_New
_swig_new_static_method = _blst.SWIG_PyStaticMethod_New

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "thisown":
            self.this.own(value)
        elif name == "this":
            set(self, name, value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


BLST_SUCCESS = _blst.BLST_SUCCESS
BLST_BAD_ENCODING = _blst.BLST_BAD_ENCODING
BLST_POINT_NOT_ON_CURVE = _blst.BLST_POINT_NOT_ON_CURVE
BLST_POINT_NOT_IN_GROUP = _blst.BLST_POINT_NOT_IN_GROUP
BLST_AGGR_TYPE_MISMATCH = _blst.BLST_AGGR_TYPE_MISMATCH
BLST_VERIFY_FAIL = _blst.BLST_VERIFY_FAIL
BLST_PK_IS_INFINITY = _blst.BLST_PK_IS_INFINITY
class SecretKey(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    keygen = _swig_new_instance_method(_blst.SecretKey_keygen)
    from_bendian = _swig_new_instance_method(_blst.SecretKey_from_bendian)
    from_lendian = _swig_new_instance_method(_blst.SecretKey_from_lendian)
    to_bendian = _swig_new_instance_method(_blst.SecretKey_to_bendian)
    to_lendian = _swig_new_instance_method(_blst.SecretKey_to_lendian)

    def __init__(self):
        _blst.SecretKey_swiginit(self, _blst.new_SecretKey())
    __swig_destroy__ = _blst.delete_SecretKey

# Register SecretKey in _blst:
_blst.SecretKey_swigregister(SecretKey)

class Scalar(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.Scalar_swiginit(self, _blst.new_Scalar(*args))
    dup = _swig_new_instance_method(_blst.Scalar_dup)
    from_bendian = _swig_new_instance_method(_blst.Scalar_from_bendian)
    from_lendian = _swig_new_instance_method(_blst.Scalar_from_lendian)
    to_bendian = _swig_new_instance_method(_blst.Scalar_to_bendian)
    to_lendian = _swig_new_instance_method(_blst.Scalar_to_lendian)
    add = _swig_new_instance_method(_blst.Scalar_add)
    sub = _swig_new_instance_method(_blst.Scalar_sub)
    mul = _swig_new_instance_method(_blst.Scalar_mul)
    inverse = _swig_new_instance_method(_blst.Scalar_inverse)
    __swig_destroy__ = _blst.delete_Scalar

# Register Scalar in _blst:
_blst.Scalar_swigregister(Scalar)

class P1_Affine(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.P1_Affine_swiginit(self, _blst.new_P1_Affine(*args))
    dup = _swig_new_instance_method(_blst.P1_Affine_dup)
    to_jacobian = _swig_new_instance_method(_blst.P1_Affine_to_jacobian)
    serialize = _swig_new_instance_method(_blst.P1_Affine_serialize)
    compress = _swig_new_instance_method(_blst.P1_Affine_compress)
    on_curve = _swig_new_instance_method(_blst.P1_Affine_on_curve)
    in_group = _swig_new_instance_method(_blst.P1_Affine_in_group)
    is_inf = _swig_new_instance_method(_blst.P1_Affine_is_inf)
    is_equal = _swig_new_instance_method(_blst.P1_Affine_is_equal)
    core_verify = _swig_new_instance_method(_blst.P1_Affine_core_verify)
    generator = _swig_new_static_method(_blst.P1_Affine_generator)
    __swig_destroy__ = _blst.delete_P1_Affine

# Register P1_Affine in _blst:
_blst.P1_Affine_swigregister(P1_Affine)
P1_Affine_generator = _blst.P1_Affine_generator

class P1(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.P1_swiginit(self, _blst.new_P1(*args))
    dup = _swig_new_instance_method(_blst.P1_dup)
    to_affine = _swig_new_instance_method(_blst.P1_to_affine)
    serialize = _swig_new_instance_method(_blst.P1_serialize)
    compress = _swig_new_instance_method(_blst.P1_compress)
    on_curve = _swig_new_instance_method(_blst.P1_on_curve)
    in_group = _swig_new_instance_method(_blst.P1_in_group)
    is_inf = _swig_new_instance_method(_blst.P1_is_inf)
    is_equal = _swig_new_instance_method(_blst.P1_is_equal)
    aggregate = _swig_new_instance_method(_blst.P1_aggregate)
    sign_with = _swig_new_instance_method(_blst.P1_sign_with)
    hash_to = _swig_new_instance_method(_blst.P1_hash_to)
    encode_to = _swig_new_instance_method(_blst.P1_encode_to)
    mult = _swig_new_instance_method(_blst.P1_mult)
    cneg = _swig_new_instance_method(_blst.P1_cneg)
    neg = _swig_new_instance_method(_blst.P1_neg)
    add = _swig_new_instance_method(_blst.P1_add)
    dbl = _swig_new_instance_method(_blst.P1_dbl)
    generator = _swig_new_static_method(_blst.P1_generator)
    __swig_destroy__ = _blst.delete_P1

# Register P1 in _blst:
_blst.P1_swigregister(P1)
P1_generator = _blst.P1_generator

class P1s(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self):
        _blst.P1s_swiginit(self, _blst.new_P1s())
    mult_pippenger = _swig_new_static_method(_blst.P1s_mult_pippenger)
    add = _swig_new_static_method(_blst.P1s_add)
    to_affine = _swig_new_static_method(_blst.P1s_to_affine)
    __swig_destroy__ = _blst.delete_P1s

# Register P1s in _blst:
_blst.P1s_swigregister(P1s)
P1s_mult_pippenger = _blst.P1s_mult_pippenger
P1s_add = _blst.P1s_add
P1s_to_affine = _blst.P1s_to_affine

class P2_Affine(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.P2_Affine_swiginit(self, _blst.new_P2_Affine(*args))
    dup = _swig_new_instance_method(_blst.P2_Affine_dup)
    to_jacobian = _swig_new_instance_method(_blst.P2_Affine_to_jacobian)
    serialize = _swig_new_instance_method(_blst.P2_Affine_serialize)
    compress = _swig_new_instance_method(_blst.P2_Affine_compress)
    on_curve = _swig_new_instance_method(_blst.P2_Affine_on_curve)
    in_group = _swig_new_instance_method(_blst.P2_Affine_in_group)
    is_inf = _swig_new_instance_method(_blst.P2_Affine_is_inf)
    is_equal = _swig_new_instance_method(_blst.P2_Affine_is_equal)
    core_verify = _swig_new_instance_method(_blst.P2_Affine_core_verify)
    generator = _swig_new_static_method(_blst.P2_Affine_generator)
    __swig_destroy__ = _blst.delete_P2_Affine

# Register P2_Affine in _blst:
_blst.P2_Affine_swigregister(P2_Affine)
P2_Affine_generator = _blst.P2_Affine_generator

class P2(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.P2_swiginit(self, _blst.new_P2(*args))
    dup = _swig_new_instance_method(_blst.P2_dup)
    to_affine = _swig_new_instance_method(_blst.P2_to_affine)
    serialize = _swig_new_instance_method(_blst.P2_serialize)
    compress = _swig_new_instance_method(_blst.P2_compress)
    on_curve = _swig_new_instance_method(_blst.P2_on_curve)
    in_group = _swig_new_instance_method(_blst.P2_in_group)
    is_inf = _swig_new_instance_method(_blst.P2_is_inf)
    is_equal = _swig_new_instance_method(_blst.P2_is_equal)
    aggregate = _swig_new_instance_method(_blst.P2_aggregate)
    sign_with = _swig_new_instance_method(_blst.P2_sign_with)
    hash_to = _swig_new_instance_method(_blst.P2_hash_to)
    encode_to = _swig_new_instance_method(_blst.P2_encode_to)
    mult = _swig_new_instance_method(_blst.P2_mult)
    cneg = _swig_new_instance_method(_blst.P2_cneg)
    neg = _swig_new_instance_method(_blst.P2_neg)
    add = _swig_new_instance_method(_blst.P2_add)
    dbl = _swig_new_instance_method(_blst.P2_dbl)
    generator = _swig_new_static_method(_blst.P2_generator)
    __swig_destroy__ = _blst.delete_P2

# Register P2 in _blst:
_blst.P2_swigregister(P2)
P2_generator = _blst.P2_generator

class P2s(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self):
        _blst.P2s_swiginit(self, _blst.new_P2s())
    mult_pippenger = _swig_new_static_method(_blst.P2s_mult_pippenger)
    add = _swig_new_static_method(_blst.P2s_add)
    to_affine = _swig_new_static_method(_blst.P2s_to_affine)
    __swig_destroy__ = _blst.delete_P2s

# Register P2s in _blst:
_blst.P2s_swigregister(P2s)
P2s_mult_pippenger = _blst.P2s_mult_pippenger
P2s_add = _blst.P2s_add
P2s_to_affine = _blst.P2s_to_affine

G1 = _blst.G1
G2 = _blst.G2
class PT(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _blst.PT_swiginit(self, _blst.new_PT(*args))
    dup = _swig_new_instance_method(_blst.PT_dup)
    is_one = _swig_new_instance_method(_blst.PT_is_one)
    is_equal = _swig_new_instance_method(_blst.PT_is_equal)
    sqr = _swig_new_instance_method(_blst.PT_sqr)
    mul = _swig_new_instance_method(_blst.PT_mul)
    final_exp = _swig_new_instance_method(_blst.PT_final_exp)
    finalverify = _swig_new_static_method(_blst.PT_finalverify)
    __swig_destroy__ = _blst.delete_PT

# Register PT in _blst:
_blst.PT_swigregister(PT)
PT_finalverify = _blst.PT_finalverify

class Pairing(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, hash_or_encode, DST):
        _blst.Pairing_swiginit(self, _blst.new_Pairing(hash_or_encode, DST))
    __swig_destroy__ = _blst.delete_Pairing
    aggregate = _swig_new_instance_method(_blst.Pairing_aggregate)
    mul_n_aggregate = _swig_new_instance_method(_blst.Pairing_mul_n_aggregate)
    commit = _swig_new_instance_method(_blst.Pairing_commit)
    merge = _swig_new_instance_method(_blst.Pairing_merge)
    finalverify = _swig_new_instance_method(_blst.Pairing_finalverify)

# Register Pairing in _blst:
_blst.Pairing_swigregister(Pairing)

cdata = _blst.cdata
memmove = _blst.memmove

cvar = _blst.cvar
BLS12_381_G1 = cvar.BLS12_381_G1
BLS12_381_NEG_G1 = cvar.BLS12_381_NEG_G1
BLS12_381_G2 = cvar.BLS12_381_G2
BLS12_381_NEG_G2 = cvar.BLS12_381_NEG_G2




from typing_extensions import TypedDict


class Operator(TypedDict):
    index: int
    active: bool
    name: str
    rewardAddress: str
    stakingLimit: int
    stoppedValidators: int
    totalSigningKeys: int
    usedSigningKeys: int


class OperatorKey(TypedDict):
    index: int
    operator_index: int
    key: bytes
    depositSignature: bytes
    used: bool



from web3 import Web3

from lido_sdk.contract import LidoContract


def get_status(w3: Web3):
    return {
        "isStopped": LidoContract.isStopped(w3)[""],
        "totalPooledEther": LidoContract.getTotalPooledEther(w3)[""],
        "withdrawalCredentials": LidoContract.getWithdrawalCredentials(w3)[""],
        "bufferedEther": LidoContract.getBufferedEther(w3)[""],
        **LidoContract.getFee(w3),
        **LidoContract.getFeeDistribution(w3),
        **LidoContract.getBeaconStat(w3),
        "last_block": w3.eth.getBlock("latest")["number"],
        "last_blocktime": w3.eth.getBlock("latest")["timestamp"],
    }



from lido_sdk.network import Network


MULTICALL_ADDRESSES = {
    Network.Mainnet: "0xeefBa1e63905eF1D7ACbA5a8513c70307C1cE441",
    Network.Kovan: "0x2cc8688C5f75E365aaEEb4ea8D6a480405A48D2A",
    Network.Rinkeby: "0x42Ad527de7d4e9d9d011aC45B31D8551f8Fe9821",
    Network.Görli: "0x77dCa2C955b15e9dE4dbBCf1246B4B85b651e50e",
    Network.xDai: "0xb5b692a88BDFc81ca69dcB1d924f59f0413A602a",
    Network.Ropsten: "0x53c43764255c17bd724f74c4ef150724ac50a3ed",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import platform
import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Extension

if sys.version_info < (3, 7):
    raise Exception("Python 3.7 or higher is required")

PLATFORMS = {"windows", "linux", "darwin", "cygwin", "android"}

# detecting target platform
target = platform.system().lower()

if "pydroid3" in sys.executable.lower():
    target = "android"

for known in PLATFORMS:
    if target.startswith(known):
        target = known

if target not in PLATFORMS:
    target = "linux"

# Package meta-data.
NAME = "lido-sdk"
DESCRIPTION = (
    "This library consolidates various functions to efficiently load network data for Lido,"
    " validate node operator keys and find key duplicates."
)
URL = "https://github.com/lidofinance/lido-python-sdk"
EMAIL = "info@lido.fi"
AUTHOR = "Lido"
REQUIRES_PYTHON = ">=3.7,<4"
VERSION = "2.1.1"


# C/C++ Extensions
LIBRARIES = []
STATIC_LIBRARIES = ["blst"]
STATIC_LIB_DIR = f"blst-lib/{target}"
LIBRARY_DIRS = ["blst-lib/"]
INCLUDE_DIRS = ["blst-lib/"]
SOURCES = [
    "blst-lib/blst_wrap.cpp",
]
DEPENDS = [
    "blst-lib/blst.h",
    "blst-lib/blst.hpp",
    "blst-lib/blst_aux.h",
    "blst-lib/libblst.a",
    "blst-lib/blstlib.dll",
]

if target == "windows":
    LIBRARIES.extend(STATIC_LIBRARIES)
    LIBRARY_DIRS.append(STATIC_LIB_DIR)
    EXTRA_OBJECTS = [
        "{}/{}.lib".format(STATIC_LIB_DIR, lib_name) for lib_name in STATIC_LIBRARIES
    ]
else:  # POSIX
    LIBRARIES.extend(STATIC_LIBRARIES)
    LIBRARY_DIRS.append(STATIC_LIB_DIR)
    EXTRA_OBJECTS = [
        "{}/lib{}.a".format(STATIC_LIB_DIR, lib_name) for lib_name in STATIC_LIBRARIES
    ]

ext_modules = [
    Extension(
        "lido_sdk.blstverify._blst",
        sources=SOURCES,
        depends=DEPENDS,
        include_dirs=INCLUDE_DIRS,
        libraries=LIBRARIES,
        library_dirs=LIBRARY_DIRS,
        extra_objects=EXTRA_OBJECTS,
        py_limited_api=True,
    ),
]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

with open("requirements.txt", "r") as file:
    requirements = [lib.strip() for lib in file.read().split("\n") if lib]

# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    package_dir={"": "."},
    packages=find_packages(exclude=("tests",)),
    package_data={"lido_sdk.contract": ["abi/*.json"]},
    # If your package is a single module, use this instead of "packages":
    # py_modules=["mypackage"],
    # entry_points={
    #     "console_scripts": ["mycli=mymodule:cli"],
    # },
    install_requires=requirements,
    tests_require=["pytest==6.2.4"],
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # How mature is this project? Common values are
        #   2 - Pre-Alpha
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Supported Python versions
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    ext_modules=ext_modules,
    dependency_links=[],
)



from .blst import PT, Pairing, P1_Affine, P2_Affine, BLST_SUCCESS

HASH_OR_ENCODE = True
DST = "BLS_SIG_BLS12381G2_XMD:SHA-256_SSWU_RO_POP_"
VALID_TYPES = (bytes, bytearray)


def verify(pubkey: bytes, message: bytes, signature: bytes) -> bool:
    if (
        (not isinstance(pubkey, VALID_TYPES))
        or (not isinstance(message, VALID_TYPES))
        or (not isinstance(signature, VALID_TYPES))
    ):
        return False

    try:
        pk_affine = P1_Affine(pubkey)
        sig_affine = P2_Affine(signature)

        ctx = Pairing(HASH_OR_ENCODE, DST)
        result = ctx.aggregate(pk_affine, sig_affine, message)
        if result != BLST_SUCCESS:
            return False

        ctx.commit()
        gtsig = PT(sig_affine)

        return ctx.finalverify(gtsig)
    except RuntimeError:
        return False



from enum import IntEnum


class Network(IntEnum):
    Mainnet = 1
    Kovan = 42
    Rinkeby = 4
    Görli = 5
    xDai = 100
    Ropsten = 3


# Network
GENESIS_FORK_VERSION = {
    Network.Mainnet: bytes.fromhex("00000000"),
    Network.Görli: bytes.fromhex("00001020"),
    Network.Ropsten: bytes.fromhex("00000000"),
}

# Existing withdrawal credentials on the chain
# Will be filtered for unique values
# Will be used as a fallback for used keys or in strict mode when validating signatures
WITHDRAWAL_CREDENTIALS = {
    Network.Mainnet: [
        "0x009690e5d4472c7c0dbdf490425d89862535d2a52fb686333f3a0a9ff5d2125e"
    ],
    Network.Görli: [
        "0x00040517ce98f81070cea20e35610a3ae23a45f0883b0b035afc5717cc2e833e"
    ],
    Network.Ropsten: [
        "0x01000000000000000000000002139137fdd974181a49268d7b0ae888634e5469",
        "0x000000000000000000000000ff139137fdd974181a49268d7b0ae888634e5469",
        "0x000000000000000000000000aa139137fdd974181a49268d7b0ae888634e5469",
        "0x010000000000000000000000aa139137fdd974181a49268d7b0ae888634e5469",
        "0x73c72beecbd832c9ce342e61a772c8cfe6f1c6d661b19a98317b5dac05ce9685",
    ],
}



from concurrent.futures import ProcessPoolExecutor
from typing import List, Tuple
from lido_sdk.eth2deposit.ssz import (
    compute_deposit_domain,
    DepositMessage,
    compute_signing_root,
)
from web3 import Web3

from lido_sdk.contract import LidoContract
from lido_sdk.methods.typing import OperatorKey
from lido_sdk.network.type import WITHDRAWAL_CREDENTIALS, GENESIS_FORK_VERSION
from lido_sdk.blstverify.verifier import verify


def find_duplicated_keys(
    keys: List[OperatorKey],
) -> List[Tuple[OperatorKey, OperatorKey]]:
    """
    Find all duplicates in list of keys

    @param keys: List of keys
    @return: Returns list of pair keys (key_1, key_2) that are duplicating
    """
    keys_dict = {}
    duplicates = []

    for key in keys:
        if key["key"] not in keys_dict:
            keys_dict[key["key"]] = key
        else:
            duplicates.append((keys_dict[key["key"]], key))

    return duplicates


def _get_withdrawal_credentials(chain_id: int):
    """
    @param chain_id: Network chain id
    @return: Possible withdrawal credentials
    """
    return [bytes.fromhex(cred[2:]) for cred in WITHDRAWAL_CREDENTIALS[chain_id]]


def validate_keys(
    w3: Web3, keys: List[OperatorKey], strict: bool = False
) -> List[OperatorKey]:
    """
    @param w3: Web3
    @param keys: List of keys to validate
    @param strict: Should be used for new keys. It will check that key was signed using contract's actual WC.
    @return: List of keys that are invalid
    """
    deposit_domain = compute_deposit_domain(GENESIS_FORK_VERSION[w3.eth.chain_id])

    actual_credential = LidoContract.getWithdrawalCredentials(w3)[""]
    possible_credentials = _get_withdrawal_credentials(w3.eth.chain_id)

    invalid_keys = []

    key_params = [
        (key, actual_credential, possible_credentials, deposit_domain, strict)
        for key in keys
    ]

    with ProcessPoolExecutor() as executor:
        for key, is_valid in zip(
            keys, executor.map(_executor_validate_key, key_params)
        ):
            if not is_valid:
                invalid_keys.append(key)

    return invalid_keys


def _executor_validate_key(data: Tuple):
    key, actual_credential, possible_credential, deposit_domain, strict = data

    is_valid = validate_key(key, actual_credential, deposit_domain)

    if not is_valid and not strict and key.get("used", False):
        for wc in possible_credential:
            is_valid = validate_key(key, wc, deposit_domain)

            if is_valid:
                break

    return is_valid


def validate_key(
    key: OperatorKey, withdrawal_credential: bytes, deposit_domain: bytes
) -> bool:
    """
    @param key: Key to check.
    @param withdrawal_credential: Possible credential.
    @param deposit_domain: Magic bytes.
    @return: Bool - Valid or Invalid this key
    """
    pub_key = key["key"]
    signature = key["depositSignature"]

    ETH32 = 32 * 10 ** 9
    deposit_message = DepositMessage(
        pubkey=key["key"],
        withdrawal_credentials=withdrawal_credential,
        amount=ETH32,
    )

    message = compute_signing_root(deposit_message, deposit_domain)
    is_valid = verify(pub_key, message, signature)

    if is_valid:
        return is_valid

    return False
